#! /usr/bin/perl

#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


#  Application: finddups.pl
#  Version:     0.5
#  Author:	R. W. Rodolico (rodo@dailydata.net)
#  Date:	20051025
#  Copyright:	Daily Data, Inc., 2005
#  Dependancies:
#	finddups.conf contains SQL queries and statements used by program
#	program relies on cksum, a Unix utility
#	program has hard coded directory separators.
#	Program also assumes a MySQL instance exists and is accessable by
#	the user running the program.
#	Because of this, the program only runs under Unix
#
#  To Do:	Handle case of directories that contain subdirectories for reporting
#		Create directory tree mirroring original tree and move files there instead of deleting
#		Make program OS independant
#		Add rescan option
#
#  Usage:	finddups.pl database [rootdir]
#
#  Description:
#
#	If rootdir is given:
#		Program will recursively descend directory under "rootdir", file name, path, file size
# 		and MD5 checksums into the table "files" in database passed in as a parameter. The
# 		"files" table is dropped and recreated each time.
# 	In all cases:
# 		1. table "dups" is created containing all checksums that occur more than once in the files
# 		   table.
# 		2. table "directories" is created containing all directories and the sum of the checksums
# 		   of all files in the directory, in essence creating a "checksum" of a directory
# 		3. table "dupdirs" is created containing a list of checksums that occur in more than one
# 		   directory
# 	If duplicate directories are detected:
# 		bash script removedups.sh is created. In this script, the command to remove all directories
# 		whose checksums are found in the "dupdirs" table are entered, with one, chosen (poorly)
# 		by the program commented out (ie, the one to be kept)
#         If no duplicate directories are found:
# 		bash script removedups.sh is created. In this script, a command to remove each individual
# 		file whose checksum is located in the "dups" table is created, with one, chosen by the
#                 program commented out (ie, the one to be kept)
#	Note: after any action is taken on the file system, you must re-load the database
#
#  NOTICE: If you run the removedups.sh script without even looking at it, you are a total idiot.
#          I wrote this HOPING that at least one copy of each file would be saved, but I won't
#          guarantee even that. It may just decide to wax your whole hard drive. The script
#	   is designed to be a guideline, to help you when you have several thousand duplicate
#          files. But, if you run the script and it gives you back a nice, shiny clean directory
#          tree with nothing in it, all I want to hear is "there's a bug in the program" and
#          some code to fix it. YOU HAVE BEEN WARNED.

# Revisions
# 0.5 RWR 20110304
# Put database definition into the conf file, and also added a variable in there to ignore files smaller than
# $MIN_PROCESS_SIZE if that is defined (set it to zero to turn that off). Placed that var definition into the
# conf file also.



BEGIN {  # I do this so I can find finddups.conf in the current directory
   $pwd = qx/pwd/;
   chomp $pwd;
   push @INC, $pwd;
}

use DBI;	    # we only use MySQL right now
# use Digest::MD5;  # for the future so we're not dependant on cksum


require "finddups.conf"; # all the queries are stored here
# Above conf file has the following definitions, plus a bunch of queries
# my $dbName;
# my $dbUser;
# my $dbPass;


my $dbh;       # using this as a global. It will hold the database handle
my $pathSeparator = '/'; # so that I hopefully will make this os independant in the future
my $scriptFileName = 'removeduplicates.sh'; # generated script to destroy your hard disk
my $currentRootDir = 'null'; # will hold the top subdirectory of the current check, which in this case is the client

# users create the stupidest file names. escape any chars except a-z, 0-9, - and _ (and the dir separator)
sub escapeFileName {
   my $filename = shift;
   $filename =~ s/([^a-zA-Z.0-9\-_~\/])/\\$1/gi;
   return $filename;
}


# simple function. Drops a table if it exists, then creates it
sub makeTable {
   my ( $tableName, $query ) = @_;
  $dbh->do( "drop table if exists $tableName" ) or die "Failed to remove old $tableName table: " . $DBI::errstr ;
  $dbh->do( $query ) or die ( "Failed to build $tableName: " . $DBI::errstr );
}

# the workhorse. Gets the checksum for a file, stores it, the path, the filename and size and
# stores it in the database
sub processFile {
   my ( $directory,$filename ) = @_;
   my $workingFileName = &escapeFileName( $directory . $filename );
   my ($disksize,$junk2) = split( ' ',`du -b $workingFileName` );
   if ($MIN_PROCESS_SIZE && $disksize < $MIN_PROCESS_SIZE) { # do not process files smaller than MIN_PROCESS_SIZE if it has been defined
      return;
   }
   #print "Working File Name =>[$workingFileName]\n";
   $directory = $dbh->quote( $directory );
   $filename = $dbh->quote( $filename );
   #my $output = `cksum $filename`;
   my ($checksum,$filesize,$junk) = split( ' ', `cksum $workingFileName` );
   my ($disksize,$junk2) = split( ' ',`du -b $workingFileName` );
   $dbh->do("insert into files(client,checksum,filesize,path,filename, disksize) values ( $currentRootDir, $checksum, $filesize, $directory,$filename, $disksize)" ) or warn ($DBI::errstr);
}

# get all entries in a directory. Process the regular files, call itself to process all subdirs, ignore anything else
sub getDirEntries {
   my $dirToRead = shift; # only parameter, the directory to scan
   my $firstTime = shift;
   my $filecount = 0;
   opendir(DIR, $dirToRead) || warn "can't opendir $dirToRead: $!"; # open the directory
   # get directories, ignoring the . and .. and any symbolic links
   my @directories = grep { ! /^((\.)|(\.\.))$/ && -d "$dirToRead/$_" && ! -l "$dirToRead/$_" } readdir(DIR);
   # reset directory. Should be able to do this without all the mess
   rewinddir DIR;
   # get files
   my @files = grep { -f "$dirToRead/$_" } readdir(DIR);
   closedir DIR;
   # let user know what is going on
   print "Processing directory $dirToRead . . " . scalar ( @files ) . " files\n";
   # for each file, call processFile
   for ( my $i = 0; $i < @files; $i++ ) {
      &processFile( $dirToRead, $files[$i] );
      $filecount++;
   }
   # now, for each directory, just call this same routine. I LOVE RECURSION
   foreach my $dir (@directories) {
      if ( $firstTime ) { $currentRootDir = $dbh->quote($dir) }
      $filecount += getDirEntries( $dirToRead . $dir . '/', 0 );
   }
   return $filecount; # returns total file count for this directory tree
}

# simply a wrapper to start the recursive director scan and creating the directories table
sub readAndProcessFiles {
  my $rootDir = shift;
  # drop and recreate the table if it exists.
  &makeTable( "files",$MAKEFILESTABLE );
  print "Scanning $rootDir, Please be patient, this can take a while\n";
  my $fileCount = &getDirEntries($rootDir, 1);
  print "Creating directories table\n";
  &makeTable( "directories", $MAKEDIRS );
  return $fileCount;
}

# simply call the queries to make the dups and dupdirs tables
sub findDups {
  print "Finding duplicate files\n";
  &makeTable( "dups",$MAKEDUPSSQL );
  print "finding duplicate directories\n";
  &makeTable( "dupdirs", $MAKEDUPDIRS );
}

# Figure out which directory to keep and which to throw away from a group of directories passed in
# not much logic to it, just plan on saving the first one, then go through list and see if one
# looks better.
# returns a list of lines of the form "rm -fRv filename for all filenames passed in
# One line in this list will have the comment character in front of it (in essence keeping that file)
sub findDirToKeep {
   my @duplicate = @_;
   my $keeping = 0; # assume we will keep the first one
   my $output = '';
   for ( my $i = 1; $i < @duplicate; $i++ ) {
      # first, if one is mixed case and the other is lower case, keep the one all lower case
      if ( ( lc $duplicate[$keep] ne $duplicate[$keep] ) and ( lc $duplicate[$i] eq $duplicate[$i] ) ) {
        $keep = $i;
        # failing that, if this one has a physically shorter length, keep it instead of the other
      } elsif ( length( $duplicate[$i] ) < length( $duplicates[$keep] )) {
         $keep = $i;
      }
   }
   # generate lines removing duplicate files
   for ( my $i = 0; $i < @duplicate; $i++ ) {
      $duplicates[$i] = &escapeFileName( $duplicates[$i] );
      $output .= '#' if $i == $keep; # keep this one
      $output .= "mv  $duplicate[$i] /home/save/\n";
   }
   return $output;
}

# Go through all directories whose checksum is in dupdirs table. Generate a script to remove all
# but one of each group
sub removeDupDirs {
  my $sth = $dbh->prepare( $SELECTDUPDIRS ) or die( "Error preparing query\n$SELECTDUPDIRS\n" . $DBI::errstr );
  $sth->execute or die( "Error executing query\n$SELECTDUPDIRS\n" . $DBI::errstr );
  my $output = '';
  my $count = 0;
  my $workingon = 0;  # tracks checksum of group we are working on
  my @duplicate;
  while ( my $data = $sth->fetchrow_hashref() ) { # just keep getting rows from table
     if ( $workingon == $$data{'checksum'} ) {    # this is just one more of the current group
        push @duplicate, $$data{'path'};
     } else { # We have a list of duplicated directories. now, we need to process it
        $output .= "\n\n# $workingon\n" if $workingon; # workingon could be zero if this is the first time we've done this
        $output .= &findDirToKeep( @duplicate );
        $count++ if $workingon;		# just for fun
        $workingon = $$data{'checksum'}; # prime with new group's checksum
        @duplicate = ($$data{'path'});   # and the current entry
     }
   }
   # stupid beginning programmer trick. I have one last group left, so I'll process it here. Should be done above
   if ( scalar( @duplicate ) ) {
        $output .= "\n\n# $workingon\n" if $workingon;
        $output .= &findDirToKeep( @duplicate );
        $count++;
  }
   print "$count directory duplications\n";
   return $output;
}


# simply scan all files and remove duplicates. I need to comment the code better.
sub removeDupFiles {
  my $sth = $dbh->prepare( $SELECTDUPS ) or die( "Error preparing query\n$SELECTDUPS\n" . $DBI::errstr );
  $sth->execute or die( "Error executing query\n$SELECTDUPS\n" . $DBI::errstr );

  my $workingon = 0;
  my @duplicate;
  $output = '';


  while ( my $data = $sth->fetchrow_hashref() ) {
     if ( $workingon == $$data{'checksum'} ) {
        push @duplicate, $$data{'path'} . "\t" .  $$data{'filename'};
     } else { # do this if we have a list of duplicates
        my $keeping = 0; # assume we will keep the first one
        my ($keepingPath, $keepingFilename ) = split( "\t", $duplicate[$keep] );
        $output .= "\n\n# $workingon\n";
        for ( my $i = 1; $i < @duplicate; $i++ ) {
           my ($path,$filename) = split( "\t", $duplicate[$i] );
           # first, if we are in the same directory and one is mixed case and the other is not, keep the one not in mixed case
           if ( ( $path eq $keepingPath ) and ( lc $filename eq $filename ) and ( $keepFilename ne lc $keepFilename ) ) {
              $keep = $i;
              ($keepingPath, $keepingFilename ) = split( "\t", $duplicate[$keep] );
           # failing that, if this one has a physically shorter path length, keep it instead of the other
           } elsif ( length( $keepingPath ) > length( $path )) {
              $keep = $i;
              ($keepingPath, $keepingFilename ) = split( "\t", $duplicate[$keep] );
           }
        }
        for ( my $i = 0; $i < @duplicate; $i++ ) {
           $duplicate[$i] =~ s/\t//gi;
           $duplicate[$i] = &escapeFileName( $duplicate[$i] );
           $output .= '#' if $i == $keep;
           $output .= "rm $duplicate[$i]\n";
        }
        $workingon = $$data{'checksum'};
        @duplicate = ($$data{'path'} . "\t" .  $$data{'filename'});
     }
   }
   return $output;
}

# routine that actually creates the remove script
sub createScripts {
  my $output =   &removeDupDirs; # first we'll see if there are any directories to remove
  $output = &removeDupFiles unless $output; # if there weren't, just remove duplicate files
  open SCRIPT, ">$scriptFileName" or die "Could not open $scriptFileName for writing";
  print SCRIPT "#! /bin/bash\n\n# Auto-generated script to remove duplicate files\n#USE AT YOUR OWN RISK -- review this script carefully\n";
  print SCRIPT "#THIS SCRIPT WILL PERMENANTLY REMOVE ENTIRE DIRECTORIES AND FILES FROM YOUR SYSTEM WITHOUT ASKING YOU FIRST\n\n";
  print SCRIPT $output;
  close SCRIPT;
}

# get parameters
my $rootDir = shift;


$rootDir .= '/' unless $rootDir =~ m/.*\/$/;

die "Usage: finddups.pl [rootdir]" unless $dbName;

# open database directory
$dbh = DBI->connect( $dbName, $dbUser , $dbPass ) or die "Could not attach to database $dbName";

# if they passed in a rootdir, we will drop the files table and recreate it
if ( $rootDir ) {
   print &readAndProcessFiles( $rootDir ) . " files scanned\n" ;
} else {
   print "No directory given, reprocessing old scan\n";
}

# find the dups
&findDups;
# create the script
#&createScripts;

1;
